##############################
## Base
###############################

##' Accessor for cellData \code{data.frame}
##'
##' @param sc An object with \code{cellData}
##' @return \code{data.frame}
##' @details \code{cData(sc)}: Return the \code{cellData} \code{data.frame}.
##' @export
setGeneric('cData', function(sc) standardGeneric('cData'))

##' Do two objects conform in dimension and type?
##'
##' Returns false if:
##' other does not inherit from Matrix
##' Otherwise returns a bitmask, which is the sum of:
##' 1: objects have same number of rows
##' 2: objects have same number of columns
##' 4: objects have same number of layers
##' @param dl DataLayer
##' @param other Another object
##' @return bitmask containing number of dimensions that agree
setGeneric('conform', function(dl, other) standardGeneric('conform'))


##' How many layers does a DataLayer have?
##'
##' @param x DataLayer
##' @return numeric, number of layers
##' @export nlayer
##' @seealso addlayer
setGeneric('nlayer', function(x) standardGeneric('nlayer'))


##' Which layer is active?
##'
##' @param x DataLayer
##' @return numeric, active layer
##' @export layer
##' @seealso addlayer
setGeneric('layer', function(x) standardGeneric('layer'))

##' Set active layer
##'
##' @param x DataLayer
##' @param value identifier of layer
##' @return DataLayer with new active layer
##' @export layer<-
##' @seealso addlayer
setGeneric('layer<-', function(x, value) standardGeneric('layer<-'))

##' Layer management methods
##'
##' A layer is a "slice" of a 3d array.  \code{SingleCellAssay} objects are stateful, in that a layer can be set, and then various routines will operate on that layer.
##' @section Methods:
##' @param x a \code{SingleCellAssay} or derived
##' @param name a \code{code} character
##' @export
##' @examples
##' data(vbetaFA)
##' vbetaFA <- addlayer(vbetaFA, 'expressed')
##' layername(vbetaFA)
##' ee <- exprs(vbetaFA)
##' layer(vbetaFA) <- 'expressed'
##' exprs(vbetaFA) <- (ee>0)*1
##' layer(vbetaFA) <- 'Et'
##' layername(vbetaFA) <- 'et'
##' layername(vbetaFA)
##' nlayer(vbetaFA)
##' @export
##' @rdname addlayer
setGeneric('addlayer', function(x, name) standardGeneric('addlayer'))

##' Get name of active layer
##'
##' @param x DataLayer 
##' @return character
##' @export layername
##' @seealso addlayer
setGeneric('layername', function(x) standardGeneric('layername'))

##' Set name of active layer
##'
##' @param x DataLayer
##' @param value character
##' @return DataLayer
##' @export
##' @seealso addlayer
setGeneric('layername<-', function(x, value) standardGeneric('layername<-'))

##' Accessor for wellKey
##'
##' This returns the wellKey, which is a unique identifier generated by \code{idvars} in the \code{mapping}
##' @param sc An object with a \code{wellKey}
##' @return \code{integer} giving the unique id generated
##' @aliases getwellKey,SingleCellAssay-method
##' @export
setGeneric('getwellKey', function(sc) standardGeneric('getwellKey'))

##' Accessor for cellData \code{AnnotatedDataFrame}
##'
##' Returns the \code{cellData}.
##' @param sc An object with \code{cellData}
##' @return \code{AnnotatedDataFrame}
##' @aliases cellData,SingleCellAssay-method
##' @export
setGeneric('cellData', function(sc) standardGeneric('cellData'))

##' @export
##' @rdname cData
##' @param value Replacement, one of \code{AnnotatedDataFrame} or \code{data.frame}
##' @details \code{cData(sc)<-value}: Replace the cellData with \code{value}, which can be either an \code{AnnotatedDataFrame} or \code{data.frame}.  The replacement is checked that it has mandatory fields defined by its class.
setGeneric("cData<-", function(sc, value) standardGeneric("cData<-"))

##' Accessor for featureData \code{data.frame}
##'
##' Returns the \code{featureData} \code{data.frame}.
##' @title fData
##' @param object An object with \code{featureData}
##' @return \code{data.frame} 
##' @docType methods
##' @rdname fData-methods
##' @name fData
##' @aliases fData,SingleCellAssay-method
##' @importMethodsFrom Biobase fData
#setGeneric('fData', function(sc) standardGeneric('fData'))
NULL

##' Accessor for featureData \code{AnnotatedDataFrame}
##'
##' Returns the \code{featureData}.
##' @param object An object with \code{featureData}
##' @return \code{AnnotatedDataFrame}
##' @docType methods
##' @rdname featureData-methods
##' @name featureData
##' @aliases featureData,SingleCellAssay-method
##' @importMethodsFrom Biobase featureData
#setGeneric('featureData', function(object) standardGeneric('featureData'))
NULL

##' @rdname subset
##' @export
setGeneric('subset')

##' @import BiocGenerics
NULL

###############################
## LmWrapper
###############################
##' fit a zero-inflated regression
##'
##' Given a design and formula, fit the zero inflated regression, storing the fits in slots
##' \code{fitC} and \code{fitD}
##' @param object inheriting from \code{LMlike}
##' @param response a vector, same length as the design, or if missing then use the current response
##' @param ... currently ignored
##' @return LMlike or subclass
##' @aliases fit,GLMlike,missing-method
##' @aliases fit,BayesGLMlike,missing-method
setGeneric('fit', function(object, response, ...) standardGeneric('fit'))

setGeneric('se.coef', function(object, ...) standardGeneric('se.coef'))

##' Coefficients of zero-inflated
##'
##' Given a fitted LMlike, return the coefficients from discrete or continuous.
##' Methods expect an argument \code{which}, a \code{character} of length one, one of "C" (continuous) or "D" (discrete) specifying which component should be returned.
##' @param object LMlike
##' @param ... passed to methods
##' @return numeric vector
##' @aliases coef,LMlike-method
##' @aliases coef,LMERlike-method
##' @aliases coef,ZlmFit-method
##' @export
setGeneric('coef', function(object, ...) standardGeneric('coef'))

##' Run a likelihood-ratio test
##'
##' Compares the change in likelihood between the current model and one subject to contrasts tested in \code{hypothesis}.
##' \code{hypothesis} can be one of a \code{character} giving complete factors or terms to be dropped from the model, \code{CoefficientHypothesis} giving names of  coefficients to be dropped, \code{Hypothesis} giving contrasts using the symbolically, or a contrast \code{matrix}, with one row for each coefficient in the full model, and one column for each contrast being tested.
##' @param object LMlike or subclass
##' @param hypothesis the hypothesis to be tested.  See details.
##' @return array giving test statistics
##' @export
##' @seealso fit
##' @seealso waldTest
##' @seealso Hypothesis
##' @seealso CoefficientHypothesis
setGeneric('lrTest', function(object, hypothesis) standardGeneric('lrTest'))

##' Run a Wald test
##'
##' Run a Wald tests on discrete and continuous components
##' \code{hypothesis} can be one of a \code{character} giving complete factors or terms to be dropped from the model, \code{CoefficientHypothesis} giving names of  coefficients to be dropped, \code{Hypothesis} giving contrasts using the symbolically, or a contrast \code{matrix}, with one row for each coefficient in the full model, and one column for each contrast being tested.
##' @param object LMlike or subclass
##' @param hypothesis the hypothesis to be tested.  See details.
##' @return array giving test statistics
##' @export
##' @seealso fit
##' @seealso lrTest
##' @seealso lht
setGeneric('waldTest', function(object, hypothesis) standardGeneric('waldTest'))

##' Variance-covariance matrix for zero inflated
##'
##' Given a fitted LMlike, return the variance-covariance from discrete or continuous.
##' \code{which}, a \code{character}, one of "C" (continuous) or "D" (discrete) must be specified, giving which component should be returned.
##' @param object LMlike
##' @param ... Additional parameters, namely \code{which}
##' @return \code{matrix}
##' @export
setGeneric('vcov', function(object) standardGeneric('vcov'))

##' Degrees of freedom of Zero inflated model
##'
##' @param object LMlike or subclass
##' @return vector giving the model degrees of freedom for continuous and discrete
##' @aliases dof,GLMlike-method
##' @aliases dof,LMERlike-method
setGeneric('dof', function(object) standardGeneric('dof'))


##' Model matrix accessor
##'
##' @param object LMlike or subclass
##' @return model.matrix if present
setGeneric('model.matrix', function(object) standardGeneric('model.matrix'))

##' Replace model matrix
##'
##' @param object LMlike or subclass
##' @param value matrix
##' @return modify object
setGeneric('model.matrix<-', function(object, value) standardGeneric('model.matrix<-'))


##' Return programmatically useful summary of a fit
##'
##' @param object LMlike or subclass
##' @return list of parameters characterizing fit
setGeneric('summarize', function(object, ...) standardGeneric('summarize'))

##' Return a human-readable summary
##'
##' @param object some object to be summarized
##' @param ... additional arguments passed to methods.
##' @return a summary, possibly printed
##' @export
setGeneric('summary', function(object, ...) standardGeneric('summary'))



###############################
## old style LRT
###############################
##' Likelihood Ratio Tests for SingleCellAssays
##'
##' Tests for a change in ET binomial proportion or mean of positive ET
##' Likelihood Ratio Test for SingleCellAssay objects
##'
##' Combined Likelihood ratio test (binomial and normal) for SingleCellAssay and derived objects.
##' This function is deprecated, please use \link{lrTest} instead.
##' @seealso zlm.SingleCellAssay, ZlmFit
##' @param ... ignored
##' @export
setGeneric("LRT",function(sca,comparison,...) standardGeneric("LRT"))

###############################
## Nanostring
###############################
##' thresholds for positive expression
##'
##' @param nsa NanostringAssay object
##' @return modified nsa
##' @export
setGeneric('thresholdNanoString', function(nsa, ...) standardGeneric('thresholdNanoString'))
